"""DB migrations commands """
from os import getcwd

from alembic.config import Config
from alembic import command


ALEMBIC_CFG = Config(f'{getcwd()}/alembic.ini')


def current(verbose=False):
    command.current(ALEMBIC_CFG, verbose=verbose)


def upgrade(revision="head"):
    command.upgrade(ALEMBIC_CFG, revision)


def downgrade(revision):
    command.downgrade(ALEMBIC_CFG, revision)



"""Settings module"""
from functools import lru_cache
from typing import Any
from os import environ
from pydantic import Json
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Brevia settings"""
    model_config = SettingsConfigDict(
        env_file='.env', extra='ignore'
    )

    verbose_mode: bool = False
    pgvector_host: str = 'localhost'
    pgvector_driver: str = 'psycopg2'
    pgvector_port: int = 5432
    pgvector_database: str = 'brevia'
    pgvector_user: str = ''
    pgvector_password: str = ''
    pgvector_pool_size: int = 10

    # Tokens
    tokens_secret: str = ''
    tokens_users: str = ''
    status_token: str = ''

    # API keys, tokens...
    openai_api_key: str = ''
    cohere_api_key: str = ''

    # Test models - only in unit tests
    use_test_models: bool = False

    # Index - textsplitter settings
    text_chunk_size: int = 2000
    text_chunk_overlap: int = 200

    # Search
    search_docs_num: int = 4

    # LLM settings
    qa_completion_llm: Json = """{
        "_type": "openai-chat",
        "model_name": "gpt-3.5-turbo-16k",
        "temperature": 0,
        "max_tokens": 1000,
        "verbose": true
    }"""
    qa_followup_llm: Json = """{
        "_type": "openai-chat",
        "model_name": "gpt-3.5-turbo-16k",
        "temperature": 0,
        "max_tokens": 200,
        "verbose": true
    }"""
    summarize_llm: Json = """{
        "_type": "openai-chat",
        "model_name": "gpt-3.5-turbo-16k",
        "temperature": 0,
        "max_tokens": 2000
    }"""

    # Embeddings
    embeddings: Json = '{"_type": "openai-embeddings"}'
    # every vendor has its own embeddings vector size, some vendors have multiple sizes
    # this value should be unique in your project to avoid calculation errors
    embeddings_size: int = 1536

    # QA
    qa_no_chat_history: bool = False  # don't load chat history
    qa_followup_sim_threshold: float = 0.735  # similitude threshold in followup
    feature_qa_lang_detect: bool = False

    # Summarization
    summ_default_chain: str = 'stuff'
    summ_token_splitter: int = 4000
    summ_token_overlap: int = 500

    # langsmith
    langchain_tracing_v2: bool = False
    langchain_endpoint: str = ''
    langchain_api_key: str = ''
    langchain_project: str = ''

    def update(
        self,
        other: dict[str, Any],
    ) -> None:
        """Update settings fields, used in unit tests"""
        for field_name in other.keys():
            key = field_name.lower()
            if hasattr(self, key):
                setattr(self, key, other.get(field_name))

    def setup_environment(self):
        """Setup some useful environment variables"""
        if self.openai_api_key:
            environ['OPENAI_API_KEY'] = self.openai_api_key
        if self.cohere_api_key:
            environ['COHERE_API_KEY'] = self.cohere_api_key
        if self.langchain_tracing_v2:
            environ['LANGCHAIN_TRACING_V2'] = "true"
            environ['LANGCHAIN_ENDPOINT'] = self.langchain_endpoint
            environ['LANGCHAIN_API_KEY'] = self.langchain_api_key
            environ['LANGCHAIN_PROJECT'] = self.langchain_project


@lru_cache
def get_settings():
    """Return Settings object instance just once (using lru_cache)"""
    settings = Settings()
    settings.setup_environment()

    return settings



"""Utilities to create langchain LLM and Chat Model instances."""
from abc import ABC, abstractmethod
from typing import Any
from langchain.llms.loading import load_llm_from_config
from langchain.llms.base import BaseLLM
from langchain.llms.fake import FakeListLLM
from langchain.chat_models.base import BaseChatModel
from langchain.chat_models import (
    ChatOpenAI,
    ChatAnthropic,
    ChatCohere,
    FakeListChatModel
)
from langchain.embeddings import OpenAIEmbeddings, CohereEmbeddings, FakeEmbeddings
from langchain.embeddings.base import Embeddings
from openai import Audio
from brevia.settings import get_settings


class FakeBreviaLLM(FakeListLLM):
    """Fake LLM for testing purposes."""
    def get_token_ids(self, text: str) -> list[int]:
        return [10] * 10


LOREM_IPSUM = """Lorem ipsum dolor sit amet, consectetur adipisici elit,
sed eiusmod tempor incidunt ut labore et dolore magna aliqua."""


def load_llm(config: dict) -> BaseLLM:
    """Load langchain LLM, use Fake LLM in test mode"""
    if test_models_in_use():
        return FakeBreviaLLM(responses=[LOREM_IPSUM] * 10)

    return load_llm_from_config(config=config)


CHAT_MODEL_TYPES: dict[str, BaseChatModel] = {
    'openai-chat': ChatOpenAI,
    'cohere-chat': ChatCohere,
    'anthropic-chat': ChatAnthropic,
    'fake-list-chat-model': FakeListChatModel,
}


class FakeBreviaChatModel(FakeListChatModel):
    """Fake LLM for testing purposes."""
    def get_token_ids(self, text: str) -> list[int]:
        return [10] * 10


def load_chatmodel(config: dict) -> BaseChatModel:
    """Load Chat Model from Config Dict."""
    if test_models_in_use():
        return FakeBreviaChatModel(responses=[LOREM_IPSUM] * 10)

    config_type = config.pop('_type')
    if config_type not in CHAT_MODEL_TYPES:
        raise ValueError(f"Loading {config_type} Chat Model not supported")

    llm_cls = CHAT_MODEL_TYPES[config_type]
    return llm_cls(**config)


class BaseAudio(ABC):
    """Base class for Audio transcription"""
    @abstractmethod
    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""


class FakeAudio(BaseAudio):
    """Test class for Audio transcription"""
    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""
        return {'text': LOREM_IPSUM}


class AudioOpenAI(BaseAudio):
    """OpenAI Audio transcription"""
    def transcribe(self, file: Any, **kwargs) -> dict:
        """transcribe Audio"""
        return Audio.transcribe(
            file=file,
            **kwargs,
        )


def load_audiotranscriber() -> BaseAudio:
    """Load Audio transcriber (only openAI supported for now)."""
    if test_models_in_use():
        return FakeAudio()

    return AudioOpenAI()


EMBEDDING_TYPES: dict[str, BaseChatModel] = {
    'openai-embeddings': OpenAIEmbeddings,
    'cohere-embeddings': CohereEmbeddings,
    'fake-embeddings': FakeEmbeddings,
}


def load_embeddings() -> Embeddings:
    """ Load Embeddings engine """
    settings = get_settings()
    if test_models_in_use():
        return FakeEmbeddings(size=settings.embeddings_size)

    config = settings.embeddings.copy()
    config_type = config.pop('_type', None)
    if config_type not in EMBEDDING_TYPES:
        raise ValueError(f'Loading "{config_type}" Embeddings not supported')

    emb_cls = EMBEDDING_TYPES[config_type]
    return emb_cls(**config)


def test_models_in_use() -> bool:
    """Check if test models are in use (via `USE_TEST_MODELS` env var)"""
    return get_settings().use_test_models



"""API endpoints definitions to handle audio input"""
from typing_extensions import Annotated
from fastapi import APIRouter, Depends
from brevia.dependencies import get_dependencies
from brevia.chat_history import get_history, ChatHistoryFilter

router = APIRouter()


@router.get('/chat_history', dependencies=get_dependencies(json_content_type=False))
def read_chat_history(filter: Annotated[ChatHistoryFilter, Depends()]):
    """ /chat_history endpoint, read stored chat history """
    return get_history(filter=filter)



"""Utility functions to run services"""
from logging import getLogger
from os import path
import yaml
from brevia.async_jobs import create_service


def run_service_from_yaml(file_path: str, num_attempts: int = 1):
    """Run service from YAML file"""
    if not path.isfile(file_path):
        raise FileNotFoundError(file_path)

    with open(file_path, 'r') as stream:
        parsed_yaml = yaml.safe_load(stream)

    attempt = 1
    while attempt <= num_attempts:
        run_service(
            service=parsed_yaml.get('service'),
            payload=parsed_yaml.get('payload'),
            output_path=parsed_yaml.get('output_path'),
            num_attempt=attempt,
        )
        attempt += 1


def save_test_result(file_path: str, num_attempt: int, output: str):
    """Save JOB result output to file, adding num attempt to file name"""
    file_name, file_extension = path.splitext(file_path)

    with open(f'{file_name}-{num_attempt}{file_extension}', 'w') as stream:
        stream.write(output)


def run_service(
    service: str,
    payload: dict,
    output_path: str = None,
    num_attempt: int = 1
):
    """Create and run service"""
    service = create_service(service)
    log = getLogger(__file__)
    attempt_message = f'Attempt: {num_attempt} - File: {payload.get("file_path")}'
    log.info('%s - start', attempt_message)
    result = service.run(payload)
    log.info('%s - end', attempt_message)

    if output_path is None:
        log.info('%s - `output_path` not specified, file not saved', attempt_message)
    else:
        save_test_result(
            file_path=output_path,
            num_attempt=num_attempt,
            output=result.get('output')
        )



"""FastAPI endpoints dependencies"""
import shutil
from pathlib import Path
import tempfile
from jose import JWTError
from langchain_community.vectorstores.pgembedding import CollectionStore
from fastapi import HTTPException, status, Header, Depends, UploadFile
from fastapi.security import OAuth2PasswordBearer
from brevia import collections, tokens
from brevia.settings import get_settings

oauth2_scheme = OAuth2PasswordBearer(tokenUrl='token')  # use token authentication


def get_dependencies(json_content_type: bool = True) -> list[Depends]:
    """Get endpoint dependencies"""
    deps = []
    # add authorization header check only if access tokens are defined
    if get_settings().tokens_secret:
        deps.append(Depends(token_auth))

    if json_content_type:
        deps.append(Depends(application_json))

    return deps


def application_json(content_type: str = Header(...)):
    """Require request MIME-type to be application/json"""

    if content_type != "application/json":
        raise HTTPException(
            status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            f"Unsupported media type: {content_type}."
            " It must be application/json",
        )


def token_auth(token: str = Depends(oauth2_scheme)):
    """Check authorization header bearer token"""
    try:
        tokens.verify_token(token)
    except JWTError as exc:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail='Invalid token',
        ) from exc


def check_collection_name(name: str) -> CollectionStore:
    """Raise a 404 response if a collection name does not exist"""
    collection = collections.single_collection_by_name(name)
    if not collection:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Collection name '{name}' was not found",
        )

    return collection


def check_collection_uuid(uuid: str):
    """Raise a 404 response if a collection uuid does not exist"""
    if not collections.collection_exists(uuid=uuid):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Collection id '{uuid}' was not found",
        )


def check_collection_name_absent(name: str):
    """Raise a 409 conflict if a collection name already exists"""
    if collections.collection_name_exists(name=name):
        raise HTTPException(
            status.HTTP_409_CONFLICT,
            f"Collection '{name}' exists",
        )


def save_upload_file_tmp(upload_file: UploadFile) -> str:
    """ Save uploaded file to temp file, return path """
    try:
        suffix = Path(upload_file.filename).suffix
        with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmp:
            shutil.copyfileobj(upload_file.file, tmp)
    finally:
        upload_file.file.close()

    return str(Path(tmp.name))



""" Callback class to handle conversation chain events """
from typing import Dict, Any, List, Sequence, Union
from uuid import UUID
import asyncio
import logging
import json
from langchain.callbacks.base import BaseCallbackHandler, AsyncCallbackHandler
from langchain.docstore.document import Document
from langchain.schema import BaseMessage
from langchain.schema.output import LLMResult


class ConversationCallbackHandler(AsyncCallbackHandler):
    """Call back handler to return conversation chain results"""
    result: Dict[str, Any] = {}
    chain_ended = asyncio.Event()

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        self.chain_ended.clear()

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        self.result = outputs
        self.chain_ended.set()

    async def wait_conversation_done(self):
        """Wait for conversation end"""
        await self.chain_ended.wait()  # await until event would be .set()

    def chain_result(self) -> str:
        """Return chain result"""
        if 'source_documents' not in self.result:
            return ''

        docs = []
        for doc in self.result['source_documents']:
            docs.append({'page_content': doc.page_content, 'metadata': doc.metadata})

        return json.dumps(docs)


class AsyncLoggingCallbackHandler(AsyncCallbackHandler):
    """Callback handler to handle logging in async calls"""
    log = None

    def __init__(self):
        self.log = logging.getLogger(__name__)
        super().__init__()

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - messages: %s", item, messages)

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - inputs: %s", item, inputs)

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - outputs: %s", item, outputs)

    async def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever start."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - query: %s", item, query)
        self.log.info("%s - serialized: %s", item, serialized)

    async def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever end."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - documents: %s", item, documents)

    async def on_retriever_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever error."""
        item = f'{run_id} / {parent_run_id}'
        self.log.error("%s - %s", item, error)

    # async def on_text(
    #     self,
    #     text: str,
    #     *,
    #     run_id: UUID,
    #     parent_run_id: UUID | None = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     """Run on arbitrary text."""
        # self.log.info('on_text')
        # self.log.info(text)


class LoggingCallbackHandler(BaseCallbackHandler):
    """Callback handler to handle logging in async calls"""
    log = None

    def __init__(self):
        self.log = logging.getLogger(__name__)
        super().__init__()

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - messages: %s", item, messages)

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - inputs: %s", item, inputs)

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - outputs: %s", item, outputs)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - prompts: %s", item, prompts)

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - response: %s", item, response)

    # def on_text(
    #     self,
    #     text: str,
    #     *,
    #     run_id: UUID,
    #     parent_run_id: UUID | None = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     """Run on arbitrary text."""
    #     self.log.info('on_text')
    #     self.log.info(text)



"""API endpoints definitions to handle document analysis like summarization"""
from typing import Annotated
from base64 import b64decode
from pathlib import Path
import tempfile
import json
import logging
from pydantic import BaseModel
from fastapi import (
    APIRouter,
    BackgroundTasks,
    HTTPException,
    status,
    UploadFile,
    Form,
)
from brevia.dependencies import get_dependencies, save_upload_file_tmp
from brevia import async_jobs, load_file
from brevia.services import SummarizeTextService

router = APIRouter()


class SummarizeBody(BaseModel):
    """ summarize input """
    text: str
    chain_type: str | None = None
    initial_prompt: dict | None = None
    iteration_prompt: dict | None = None
    token_data: bool = False


@router.post('/summarize', dependencies=get_dependencies())
def sum_documents(summarize: SummarizeBody):
    """
    /summarize endpoint:
    Summarize text and return a summary with algorithm selection
    and custom prompt options.

    Args:
        summarize (SummarizeBody): An object representing the request data
        It contains the following parameters:
            text (str): The text to be summarized
            chain_type: The main langchain summarization chain type should be one of
                "stuff", "map_reduce", and "refine".
                if not providerd stuff is used by default
            initial_prompt: Optional custom prompt to be used in the selected langchain
                chain type to replace the main chain prompt defaults
            iteration_prompt: Optional custom prompts to be used in the selected
                langchain chain type to replace the second chain promopt defaults
            token_data (bool): A boolean indicating whether to include
                token-level data in the summary

    Returns:
        JSON object representing the summary of the provided text

    Raises:
        - HTTPException with status code 400 if the 'text' field is empty

    With this endpoint, you can choose the summarization algorithm from those available
    in the Langchain library and customize the prompts according on the chosen algorithm
    """
    summarize.text = load_file.cleanup_text(summarize.text).strip()
    if not summarize.text:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'Empty text field',
        )

    service = SummarizeTextService()
    return service.run(summarize.model_dump())


@router.post(
    '/upload_summarize',
    dependencies=get_dependencies(json_content_type=False)
)
def upload_summarize(
    background_tasks: BackgroundTasks,
    chain_type: Annotated[str, Form()] = '',
    initial_prompt: Annotated[str, Form()] = '',
    iteration_prompt: Annotated[str, Form()] = '',
    file: UploadFile | None = None,
    file_content: Annotated[str, Form()] = '',
    token_data: Annotated[bool, Form()] = False,
    payload: Annotated[str, Form()] = '{}',
):
    """
    Upload a PDF file and perform summarization with algorithm
    selection and custom prompt options.

    Args:
        background_tasks (BackgroundTasks): A background task manager
            for handling asynchronous tasks
        chain_type (str, optional): The main Langchain summarization chain type
            Should be one of "stuff", "map_reduce", or "refine"
            If not provided, 'stuff' is used by default
        initial_prompt (str, optional): Optional custom prompt as json string to be
            used in the selected Langchain chain type to replace the main chain
            prompt defaults
        iteration_prompt (str, optional): Optional custom prompt as json string
            to be used in the selected Langchain chain type to replace the second
            chain prompt defaults
        file (UploadFile | None): An uploaded PDF file to be summarized
        file_content (str, optional): Content of the PDF file provided
            as a base64-encoded string
        token_data (bool, optional): A boolean indicating whether to include
            token-level data in the summary
        payload (str, optional): Optional payload in JSON format to use in the async job
            service to add custom options and custom fields

    Returns:
        A JSON object representing the job UUID for the asynchronous summarization task

    Raises:
        HTTPException with status code 400 if either 'file' or 'file_content'
        form field is missing

    With this endpoint, you can upload a PDF file for summarization, select a
    summarization algorithm from those available in the Langchain library,
    and customize the prompts according to your chosen algorithm
    """
    if file:
        log = logging.getLogger(__name__)
        log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
        log.info(
            "Type '%s' - initial prompt: '%s' - iteration prompt: '%s'",
            chain_type,
            initial_prompt,
            iteration_prompt
        )
        tmp_path = save_upload_file_tmp(file)
    elif file_content:
        tmp_path = save_base64_tmp_file(file_content=file_content)
    else:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'One of "file" or "file_content" form field is mandatory',
        )

    job = async_jobs.create_job(
        service='brevia.services.SummarizeFileService',
        payload=json.loads(payload) | {
            'file_path': tmp_path,
            'chain_type': chain_type,
            'initial_prompt':
                json.loads(initial_prompt) if initial_prompt else None,
            'iteration_prompt':
                json.loads(iteration_prompt) if iteration_prompt else None,
            'token_data': token_data,
        }
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


@router.post('/upload_analyze', dependencies=get_dependencies(json_content_type=False))
def upload_analyze(
    file: UploadFile,
    service: Annotated[str, Form()],
    background_tasks: BackgroundTasks,
    payload: Annotated[str, Form()] = '{}',
):
    """
    Upload a file and perform some analysis using a `service` class
    """
    log = logging.getLogger(__name__)
    log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
    tmp_path = save_upload_file_tmp(file)

    payload = json.loads(payload)
    payload['file_path'] = tmp_path
    job = async_jobs.create_job(
        service=service,
        payload=payload,
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


def save_base64_tmp_file(file_content: str) -> str:
    """ Save base64 file content to temp file, return path """

    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        content_as_bytes = str.encode(file_content)  # convert string to bytes
        content_recovered = b64decode(content_as_bytes)  # decode base64string
        with open(tmp.name, 'wb') as file:
            file.write(content_recovered)

    return str(Path(tmp.name))



"""API endpoints definitions for indexing actions"""
from typing import Annotated
from os import path
import json
import logging
from pydantic import BaseModel
from fastapi import APIRouter, HTTPException, status, UploadFile, Form
from langchain.docstore.document import Document
from langchain_community.vectorstores.pgembedding import CollectionStore
from brevia.dependencies import get_dependencies, save_upload_file_tmp
from brevia import index, collections, load_file

router = APIRouter()


class IndexBody(BaseModel):
    """ /index request body """
    content: str
    collection_id: str
    document_id: str
    metadata: dict = {}


@router.post('/index', status_code=204, dependencies=get_dependencies())
def index_document(item: IndexBody):
    """ Add single document to collection index """
    collection = load_collection(collection_id=item.collection_id)
    # remove same document if already indexed
    index.remove_document(
        collection_id=item.collection_id,
        document_id=item.document_id,
    )
    index.add_document(
        document=Document(page_content=item.content, metadata=item.metadata),
        collection_name=collection.name,
        document_id=item.document_id,
    )


def load_collection(collection_id: str) -> CollectionStore:
    """ Load collection by ID and throw 404 if not found"""
    collection = collections.single_collection(collection_id)
    if collection is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Collection id '{collection_id}' was not found",
        )
    return collection


@router.post(
    '/index/upload',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False)
)
def upload_and_index(
    file: UploadFile,
    collection_id: Annotated[str, Form()],
    document_id: Annotated[str, Form()],
    metadata: Annotated[str | None, Form()] = None,
):
    """
    Upload a PDF file and perform index on a collection
    """
    collection = load_collection(collection_id=collection_id)
    log = logging.getLogger(__name__)
    log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
    log.info("Collection '%s' - document %s", collection.name, document_id)
    if file.content_type not in ['application/pdf', 'text/plain']:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            f'Unsupported file content type "{file.content_type}"',
        )
    tmp_path = save_upload_file_tmp(file)
    if metadata is not None:
        metadata = json.loads(metadata)
    else:
        metadata = {'source': path.basename(tmp_path)}

    # remove same document if already indexed
    index.remove_document(
        collection_id=collection_id,
        document_id=document_id,
    )

    text = load_file.read(file_path=tmp_path)
    index.add_document(
        document=Document(
            page_content=text,
            metadata=metadata,
        ),
        collection_name=collection.name,
        document_id=document_id,
    )


@router.delete(
    '/index/{collection_id}/{document_id}',
    status_code=204,
    dependencies=get_dependencies(json_content_type=False)
)
def remove_document(collection_id: str, document_id: str):
    """ Remove document from collection index """
    index.remove_document(
        collection_id=collection_id,
        document_id=document_id,
    )


@router.get(
    '/index/{collection_id}/{document_id}',
    dependencies=get_dependencies(json_content_type=False)
)
def read_document(collection_id: str, document_id: str):
    """ Read document from collection index """
    return index.read_document(
        collection_id=collection_id,
        document_id=document_id,
    )



"""API endpoints for question answering and search"""
from typing import Annotated
import asyncio
from langchain.callbacks import AsyncIteratorCallbackHandler, get_openai_callback
from langchain.callbacks.openai_info import OpenAICallbackHandler
from langchain.chains.base import Chain
from fastapi import APIRouter, Header
from fastapi.responses import StreamingResponse
from brevia import chat_history
from brevia.dependencies import (
    get_dependencies,
    check_collection_name,
)
from brevia.callback import ConversationCallbackHandler
from brevia.language import Detector
from brevia.query import SearchQuery, ChatParams, conversation_chain, search_vector_qa
from brevia.models import test_models_in_use

router = APIRouter()


class ChatBody(ChatParams):
    """ /chat request body """
    question: str
    collection: str
    chat_history: list = []
    chat_lang: str | None = None
    token_data: bool = False


@router.post('/prompt', dependencies=get_dependencies(), deprecated=True)
@router.post('/chat', dependencies=get_dependencies())
async def chat_action(
    chat_body: ChatBody,
    x_chat_session: Annotated[str | None, Header()] = None,
):
    """ /chat endpoint, ask chatbot about a collection of documents """
    collection = check_collection_name(chat_body.collection)
    if not collection.cmetadata:
        collection.cmetadata = {}
    lang = chat_language(chat_body=chat_body, cmetadata=collection.cmetadata)

    conversation_handler = ConversationCallbackHandler()
    stream_handler = AsyncIteratorCallbackHandler()
    chain = conversation_chain(
        collection=collection,
        chat_params=ChatParams(**chat_body.model_dump()),
        answer_callbacks=[stream_handler] if chat_body.streaming else [],
        conversation_callbacks=[conversation_handler]
    )

    if not chat_body.streaming or test_models_in_use():
        return await run_chain(
            chain=chain,
            chat_body=chat_body,
            lang=lang,
            x_chat_session=x_chat_session,
        )

    asyncio.create_task(run_chain(
        chain=chain,
        chat_body=chat_body,
        lang=lang,
        x_chat_session=x_chat_session,
    ))

    async def event_generator(
        stream_callback: AsyncIteratorCallbackHandler,
        conversation_callback: ConversationCallbackHandler,
        source_docs: bool = False,
    ):
        ait = stream_callback.aiter()

        async for token in ait:
            yield token

        if not source_docs:
            yield ''
        else:
            await conversation_callback.wait_conversation_done()

            yield conversation_callback.chain_result()

    return StreamingResponse(event_generator(
        stream_callback=stream_handler,
        conversation_callback=conversation_handler,
        source_docs=chat_body.source_docs,
    ))


def chat_language(chat_body: ChatBody, cmetadata: dict) -> str:
    """Retrieve the language to be used in Q/A response"""
    chat_lang = chat_body.chat_lang or cmetadata.get('chat_lang')
    if chat_lang:
        return chat_lang

    return Detector().detect(chat_body.question)


def retrieve_chat_history(history: list, question: str, session: str = None) -> list:
    """Retrieve chat history to be used in final prompt creation"""
    chat_hist = chat_history.history(
        chat_history=history,
        session=session,
    )
    if chat_hist and not chat_history.is_related(chat_hist, question):
        chat_hist = []

    return chat_hist


async def run_chain(
    chain: Chain,
    chat_body: ChatBody,
    lang: str,
    x_chat_session: str,
):
    """Run chain usign async methods and return result"""
    with get_openai_callback() as callb:
        result = await chain.acall({
            'question': chat_body.question,
            'chat_history': retrieve_chat_history(
                history=chat_body.chat_history,
                question=chat_body.question,
                session=x_chat_session,
            ),
            'lang': lang,
        })

    return chat_result(
        result=result,
        callb=callb,
        chat_body=chat_body,
        x_chat_session=x_chat_session
    )


def chat_result(
    result: dict,
    callb: OpenAICallbackHandler,
    chat_body: ChatBody,
    x_chat_session: str | None = None,
) -> dict:
    """ Handle chat result: save chat history and return answer """
    answer = result['answer'].strip(" \n")

    chat_history.add_history(
            session_id=x_chat_session,
            collection=chat_body.collection,
            question=chat_body.question,
            answer=answer,
            metadata=callb.__dict__,
    )

    return {
        'bot': answer,
        'docs': None if not chat_body.source_docs else result['source_documents'],
        'token_data': None if not chat_body.token_data else callb.__dict__
    }


@router.post('/search', dependencies=get_dependencies())
def search_documents(search: SearchQuery):
    """
        /search endpoint:
        Search the first {docs_num} relevant documents for a question

        Specify distance_strategy:
            EUCLIDEAN = EmbeddingStore.embedding.l2_distance (default)
            COSINE = EmbeddingStore.embedding.cosine_distance
            MAX_INNER_PRODUCT = EmbeddingStore.embedding.max_inner_product
    """
    collection = check_collection_name(search.collection)
    if search.docs_num is None and 'docs_num' in collection.cmetadata:
        search.docs_num = int(collection.cmetadata['docs_num'])
    result = search_vector_qa(search=search)

    return extract_content_score(result)


def extract_content_score(data_list) -> list:
    """Extract score from search result list."""

    if 'error' in data_list:
        return {'error': data_list['error']}
    extracted_data = []
    for item in data_list:
        content = item[0].page_content.strip(" \n")
        score = item[1]
        extracted_data.append({"content": content, "score": score})
    return extracted_data
